! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2014 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                 One-dimensional partitioning                 !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!#define FD4_PART_1D_SCAN_HAN
!#define FD4_PART_1D_SCAN_HAN_BINSEARCH

#ifndef __bg__
#ifdef FD4_PART_1D_SCAN_HAN
#warning compiling with FD4_PART_1D_SCAN_HAN
#endif
#ifdef FD4_PART_1D_SCAN_HAN_BINSEARCH
#warning compiling with FD4_PART_1D_SCAN_HAN_BINSEARCH
#endif
#endif

!! One-dimensional partitioning of prefix-summed weight arrays in partitions with
!! (almost) equal load. Used for space-filling curve partitioning.
!!
!! Preprocessor options:
!!  - $FD4_PART_1D_SCAN_HAN: use Han's method to scan for partition borders (Han et al. 1992) in H1 and H2 heuristics
!!  - $FD4_PART_1D_SCAN_HAN_BINSEARCH: use Han's method in binary search methods $fd4_part_1d_binsrch and $fd4_part_1d_parallel
!!
!! @author Matthias Lieber
module fd4_part_1d_mod

  use fd4_globaldef_mod
  use fd4_mpi_mod
  use timing_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  ! public methods
  public :: fd4_part_1d_binsrch
  public :: fd4_part_1d_parallel
  public :: fd4_part_1d_simple_h1
  public :: fd4_part_1d_simple_h2
  public :: fd4_part_1d_simple_h1_reverse
  public :: fd4_part_1d_scan
  public :: fd4_part_1d_simple_rb
  public :: fd4_part_1d_binsrch_ebs

  public :: fd4_part_1d_scan_bin
  !public :: fd4_part_1d_scan_han
  !public :: fd4_part_1d_scan_bin_reverse
  !public :: fd4_part_1d_scan_han_reverse
  
  !integer(i8k), save, public :: cnt=0
  
contains  

  !! Search the largest value which is not larger than $sumload in the array of prefix summed block weights
  !! $blkweights_ps starting at $pstart and return the index in $blkweights_ps.
  !! Returns a value >=$pstart and <=$nblocks.
  !!
  !! Improved algorithm with log(nblocks/nprocs) complexity, based on ideas proposed by Han et al. 1992.
  !!
  !! In the first call $step should be $avgblk.
  !!
  function fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, step) result(res)
    integer(i_k), intent(in)    :: nblocks                  !! number of blocks
    real(r8k),    intent(in)    :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    real(r8k),    intent(in)    :: sumload                  !! upper bound for search in $blkweights_ps values
    integer(i_k), intent(in)    :: avgblk                   !! average number of blocks per partition (nblocks/nprocs)
    integer(i_k), intent(inout) :: step                     !! start of search interval of length avgblk, use for subsequent calls
    integer(i_k) :: res
    
    ! linear search in (large) steps of size avgblk for a
    ! position step > res with (step - avgblk) <= res
    do while(blkweights_ps(step) < sumload)
      step = step + avgblk
      if(step>=nblocks) then
        step=nblocks
        exit
      end if
    end do

    ! start binary search between (step-avgblk) and step
    res = fd4_part_1d_scan_bin(nblocks, blkweights_ps, step-avgblk, step+1, sumload)

    !if(blkweights_ps(res)>sumload) stop 'error 1'
    !if(blkweights_ps(MIN(res+1,nblocks))<=sumload .and. res<nblocks) stop 'error 2'
 
  end function fd4_part_1d_scan_han



  !! Reverse search the smallest value which is not smaller than $sumload in the array of prefix summed block weights
  !! $blkweights_ps starting at $pstart and return the index in $blkweights_ps.
  !! Returns a value <=$pstart and >0.
  !!
  !! This is the right-to-left version of $fd4_part_1d_scan_han as proposed by Pinar and Aykanat.
  !!
  !! In the first call $step should be $nblocks - $avgblock.
  !!
  function fd4_part_1d_scan_han_reverse(nblocks, blkweights_ps, sumload, avgblk, step) result(res)
    integer(i_k), intent(in)    :: nblocks                  !! number of blocks
    real(r8k),    intent(in)    :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    real(r8k),    intent(in)    :: sumload                  !! upper bound for search in $blkweights_ps values
    integer(i_k), intent(in)    :: avgblk                   !! average number of blocks per partition (nblocks/nprocs)
    integer(i_k), intent(inout) :: step                     !! start of search interval of length avgblk, use for subsequent calls
    integer(i_k) :: res
    
    ! linear search in (large) steps of size avgblk for a
    ! position step < res with (step + avgblk) >= res
    do while(blkweights_ps(step) >= sumload)
      step = step - avgblk
      if(step<0) then
        step=0
        exit
      end if
    end do

    ! start binary search
    res = fd4_part_1d_scan_bin_reverse(nblocks, blkweights_ps, step, step+avgblk, sumload)
 
  end function fd4_part_1d_scan_han_reverse



  !! Search the largest value which is not larger than $sumload in the array of prefix summed block weights
  !! $blkweights_ps starting at $pstart and return the index in $blkweights_ps.
  !! Returns a value >=$pstart and <=$nblocks.
  !!
  !! Uses a more sophisticated algorithm which starts searching at index $guess.
  !! Typically faster than $fd4_part_1d_scan_han.
  !! Returns a new value for $guess for subsequent searches.
  !!
  !! Called by $fd4_part_1d_binsrch, $fd4_part_1d_parallel, $fd4_part_1d_simple_h1, and $fd4_part_1d_simple_h2.
  !!
  function fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess) result(res)
    integer(i_k), intent(in)    :: nblocks                  !! number of blocks
    real(r8k),    intent(in)    :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(in)    :: pstart                   !! start index for search
    real(r8k),    intent(in)    :: sumload                  !! upper bound for search in $blkweights_ps values
    integer(i_k), intent(inout) :: guess                    !! guess for result value for clever algorithm
    integer(i_k) :: res
    
    integer(i_k) :: i

    ! try a clever guess and then search with single steps forward or backward
    if( blkweights_ps(guess) > sumload ) then
      ! guessed too many blocks
      res = pstart
      do i=guess-1,pstart,-1
        !cnt=cnt+1
        if( blkweights_ps(i) <= sumload ) then
          res=i
          exit
        end if
      end do
    else
      ! guessed too few blocks (or exactly the right number of blocks - still need to check this)
      res = nblocks
      do i=guess+1,nblocks
        !cnt=cnt+1
        if( blkweights_ps(i) > sumload ) then
          res=i-1
          exit
        end if
      end do
    end if
    ! guess number of blocks for next rank = number of blocks for this rank
    guess=MIN(res+res-pstart,nblocks)

  end function fd4_part_1d_scan


  !! Search the largest value which is not larger than $sumload in the array of prefix summed block weights
  !! $blkweights_ps and return the index in $blkweights_ps.
  !!
  !! The search is performed between $pstart and $pend. Use $pstart=0 and $pend=$nblocks for search over all weights.
  !!
  !! Uses binary search in the $blkweights_ps array.
  !!
  function fd4_part_1d_scan_bin(nblocks, blkweights_ps, pstart, pend, sumload) result(res)
    integer(i_k), intent(in)    :: nblocks                  !! number of blocks
    real(r8k),    intent(in)    :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(in)    :: pstart                   !! start index for search
    integer(i_k), intent(in)    :: pend                     !! end index for search
    real(r8k),    intent(in)    :: sumload                  !! upper bound for search in $blkweights_ps values
    integer(i_k) :: res
    
    integer(i_k) :: p0, p1, plen, i, j

    p0=pstart
    p1=pend
    plen=pend-pstart
    do j=2,pend-pstart
      if(plen<=1) exit
      i=p0+ISHFT(plen,-1)
      if( blkweights_ps(i) > sumload ) then
        p1=i
      else
        p0=i
      end if
      plen=p1-p0
    end do
    
    res=p0

  end function fd4_part_1d_scan_bin


  !! Search the smallest value which is not smaller than $sumload in the array of prefix summed block weights
  !! $blkweights_ps and return the index in $blkweights_ps.
  !!
  !! The search is performed between $pstart and $pend. Use $pstart=0 and $pend=$nblocks for search over all weights.
  !!
  !! Uses binary search in the $blkweights_ps array.
  !!
  function fd4_part_1d_scan_bin_reverse(nblocks, blkweights_ps, pstart, pend, sumload) result(res)
    integer(i_k), intent(in)    :: nblocks                  !! number of blocks
    real(r8k),    intent(in)    :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(in)    :: pstart                   !! start index for search
    integer(i_k), intent(in)    :: pend                     !! end index for search
    real(r8k),    intent(in)    :: sumload                  !! upper bound for search in $blkweights_ps values
    integer(i_k) :: res
    
    integer(i_k) :: p0, p1, plen, i, j

    p0=pstart
    p1=pend
    plen=pend-pstart
    do j=2,pend-pstart
      if(plen<=1) exit
      i=p0+ISHFT(plen,-1)
      if( blkweights_ps(i) < sumload ) then
        p0=i
      else
        p1=i
      end if
      plen=p1-p0
    end do
    
    res=p1

  end function fd4_part_1d_scan_bin_reverse


  !! Calculate a well-balanced partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! The algorithm is exact for $eps=0.0 and large $maxsteps.
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - binary search for a bottleneck value until error <= $eps
  !!  - found bottleneck value is used to create partitioning
  !!  - it may happen that the last ranks end up having no local blocks, this is no bug
  !!
  !! Caution: Be sure that $maxweight is really the max difference of subsequent values in $blkweights_ps
  !! and not the maximum weight in the block weights before prefix sum. Small rounding errors due
  !! to limited computer precision can lead to different values. With a too small maxweight the algorithm
  !! misses the exit condition and thus runs $maxsteps steps.
  !! If the block weights do not utilize the full real(8) precision, this problem should not appear.
  !!
  !! The algorithm is similar to the method EBS (exact bisection) from:
  !!
  !! Pinar, A. and C. Aykanat: Fast optimal load balancing algorithms for 1D partitioning.
  !! Journal of Parallel and Distributed Computing, 64(8):974-996, 2004.
  !!
  subroutine fd4_part_1d_binsrch(nblocks, nprocs, blkweights_ps, maxweight, eps, maxsteps, partition, maxload, reqsteps)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    real(r8k),    intent(in)  :: maxweight                !! max weight in weight array
    real(r8k),    intent(in)  :: eps                      !! max. difference of the resulting bottleneck value from the minimal
    integer(i_k), intent(in)  :: maxsteps                 !! max. number of binary search steps, may stop earlier depending on eps
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller
    integer(i_k), intent(out) :: reqsteps                 !! required number of search steps to reach the accuracy specified by eps

    integer(i_k) :: crank, pstart, pend, step
    real(r8k) :: totalweight, targetload, lbload, ubload, avload, sumload, bottleneck
    integer(i_k) :: guess, avgblk
    real(r8k) :: minp1load ! smallest realizable bottleneck value greater then current targetload
    real(r8k) :: last_targetload ! targetload from last binary search step

    reqsteps = maxsteps
    totalweight = blkweights_ps(nblocks)
    maxload = totalweight
    targetload = -1.0_r8k
    
    avgblk = nblocks / nprocs
    avload = totalweight / nprocs   ! average proc load = optimal balance
    lbload = MAX(avload, maxweight) ! lower bound for max proc load, best value that can be reached
    ubload = avload + maxweight     ! upper bound for max proc load, worst value
    !write(*,'(A,E12.4,E12.4,A,F7.4,F7.4)') 'lb/ub: ',lbload,ubload,' ~ ',avload/lbload,avload/ubload
    
    
    !! find good approximation of minimal max proc load ("bottleneck value")

    do step=1,maxsteps

      last_targetload = targetload
      targetload = lbload * 0.5_r8k + ubload * 0.5_r8k
      if(last_targetload == targetload) then
        ! parachute in case of floating point precision issues
        reqsteps = step - 1
        exit
      end if
      guess = avgblk

      sumload = targetload
      pstart = 0
      maxload = 0
      minp1load = totalweight
      do crank=1,nprocs-1
        ! find partition for rank crank with load <= targetload
#ifdef FD4_PART_1D_SCAN_HAN_BINSEARCH
        pend = fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, guess)
#else
        pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
#endif
        maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
        if(pend==nblocks) exit
        minp1load = MIN(blkweights_ps(pend+1) - blkweights_ps(pstart), minp1load)
        pstart = pend
        sumload = targetload + blkweights_ps(pstart)
      end do ! crank
      
      maxload = MAX(blkweights_ps(nblocks) - blkweights_ps(pstart), maxload)
 
      if(sumload<totalweight) then
        ! miss
        lbload = MIN(blkweights_ps(nblocks) - blkweights_ps(pstart), minp1load)
        !ubload = MIN(ubload,maxload)
      else
        ! hit
        ubload = maxload
      end if
      
      !write(*,'(I6,E32.24,F7.4)') step,targetload,avload/targetload

      ! exit the loop if specified accuracy has been reached
      if(ubload-lbload<=eps) then
        reqsteps = step
        exit
      end if
      
    end do ! steps
    
    bottleneck = ubload

    !print *,ubload,lbload,ubload-lbload

    !! finally create partition with given maximum load ("bottleneck value")

    guess = avgblk
    pstart = 0
    sumload = bottleneck
    do crank=1,nprocs-1
      ! find partition for rank crank with load <= bottleneck
#ifdef FD4_PART_1D_SCAN_HAN_BINSEARCH
      pend = fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, guess)
#else
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
#endif
      pstart = pend
      ! set partition vector
      partition(crank) = pstart+1
      sumload = bottleneck + blkweights_ps(pstart)
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck

  end subroutine fd4_part_1d_binsrch


  !! Calculate a well-balanced partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! The algorithm is exact for $eps=0.0 and large $maxsteps.
  !!
  !! Algorithm:
  !!  - embarrassingly parallel
  !!  - every process tries a different bottleneck value interval
  !!    - first each process checks its lower bound for the bottleneck
  !!    - if this is successfull, further local checks are skipped
  !!    - else binary search is applied within the interval until error <= $eps
  !!    - $maxsteps=0 disables the search, so that each rank checks its lower bound only
  !!  - best value found is determined with MPI_Allreduce
  !!  - found bottleneck value is used to create partitioning
  !!  - it may happen that the last ranks end up having no local blocks, this is no bug
  !!
  !! Caution: Be sure that $maxweight is really the max difference of subsequent values in $blkweights_ps
  !! and not the maximum weight in the block weights before prefix sum. Small rounding errors due
  !! to limited computer precision can lead to different values. With a too small maxweight the algorithm
  !! misses the exit condition and thus runs $maxsteps steps.
  !! If the block weights do not utilize the full real(8) precision, this problem should not appear.
  !!
  !! The bisection algorithm is based on the method EBS (exact bisection) from:
  !!
  !! Pinar, A. and C. Aykanat: Fast optimal load balancing algorithms for 1D partitioning.
  !! Journal of Parallel and Distributed Computing, 64(8):974-996, 2004.
  !!
  subroutine fd4_part_1d_parallel(nblocks, nprocs, blkweights_ps, maxweight, eps, maxsteps, mpi, &
                                          partition, maxload, reqsteps, mpi_time, err)
    integer(i_k), intent(in)   :: nblocks                  !! number of blocks
    integer(i_k), intent(in)   :: nprocs                   !! number of processes
    real(r8k),    intent(in)   :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    real(r8k),    intent(in)   :: maxweight                !! max weight in weight array
    real(r8k),    intent(in)   :: eps                      !! max. difference of the resulting bottleneck value from the minimal
    integer(i_k), intent(in)   :: maxsteps                 !! max. number of binary search steps, may stop earlier depending on eps
    type(fd4_mpi),intent(in)   :: mpi                      !! domain's MPI context
    integer(i_k), intent(out)  :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out)  :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller
    integer(i_k), intent(out)  :: reqsteps                 !! required number of search steps to reach the accuracy specified by eps
    integer(i8k), intent(out)  :: mpi_time                 !! the time of this rank for MPI_Allreduce in us
    integer(i_k), intent(out)  :: err                      !! error status: 0...ok

    integer(i_k) :: crank, pstart, pend, step
    real(r8k) :: totalweight, targetload, lbload, ubload, avload, sumload, bottleneck
    real(r8k) :: mybottleneck, mylbload, myubload
    integer(i_k) :: guess, mysteps, avgblk
    real(r8k) :: minp1load ! smallest realizable bottleneck value greater then current targetload
    integer(i8k) :: timing_us0
    real(r8k) :: last_targetload ! targetload from last binary search step
    
    err = 0
    
    totalweight = blkweights_ps(nblocks)
    
    avgblk = nblocks / nprocs
    avload = totalweight / nprocs   ! average proc load = optimal balance
    lbload = MAX(avload, maxweight) ! lower bound for max proc load, best value that can be reached
    ubload = avload + maxweight     ! upper bound for max proc load, worst value
    

    !! every process starts with different [lb,ub] and checks if a partition exists with
    !! a bottleneck value within this interval

    mylbload = lbload +  mpi%rank    * (ubload-lbload) / mpi%nprocs
    myubload = lbload + (mpi%rank+1) * (ubload-lbload) / mpi%nprocs
    !write(*,'(I6,A,E12.4,E12.4,A,F7.4,F7.4)') mpi%rank,'  lb/ub: ',mylbload,myubload,' ~ ',avload/mylbload,avload/myubload


    !! find good approximation of minimal max proc load ("bottleneck value")

    !! start with checking the local mylbload, if this is valid, we can skip further search
    targetload = mylbload
    guess = avgblk
    sumload = targetload
    pstart = 0
    maxload = 0
    minp1load = totalweight
    do crank=1,nprocs-1
      ! find partition for rank crank with load <= targetload
#ifdef FD4_PART_1D_SCAN_HAN_BINSEARCH
      pend = fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, guess)
#else
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
#endif
      maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
      if(pend==nblocks) exit
      minp1load = MIN(blkweights_ps(pend+1) - blkweights_ps(pstart), minp1load)
      pstart = pend
      sumload = targetload + blkweights_ps(pstart)
    end do ! crank

    maxload = MAX(blkweights_ps(nblocks) - blkweights_ps(pstart), maxload)

    if(sumload<totalweight) then
      ! miss
      mylbload = MIN(blkweights_ps(nblocks) - blkweights_ps(pstart), minp1load)
      mybottleneck = ubload
      mysteps = maxsteps
    else
      ! hit
      mybottleneck = maxload
      mysteps = 0
    end if

    ! skip further search if specified accuracy has been reached
    if(myubload-mylbload<=eps) then
      mysteps = 0
    end if
    
    reqsteps = mysteps + 1
    targetload = -1.0_r8k

    !! now go through binary search
    do step=1,mysteps

      last_targetload = targetload
      targetload = mylbload * 0.5_r8k + myubload * 0.5_r8k
      if(last_targetload == targetload) then
        ! parachute in case floating point precision issues led us here
        reqsteps = step - 1
        exit
      end if
      guess = avgblk
      
      sumload = targetload
      pstart = 0
      maxload = 0
      minp1load = totalweight
      do crank=1,nprocs-1
        ! find partition for rank crank with load <= targetload
#ifdef FD4_PART_1D_SCAN_HAN_BINSEARCH
        pend = fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, guess)
#else
        pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
#endif
        maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
        if(pend==nblocks) exit
        minp1load = MIN(blkweights_ps(pend+1) - blkweights_ps(pstart), minp1load)
        pstart = pend
        sumload = targetload + blkweights_ps(pstart)
      end do ! crank
 
      maxload = MAX(blkweights_ps(nblocks) - blkweights_ps(pstart), maxload)

      if(sumload<totalweight) then
        ! miss
        mylbload = MIN(blkweights_ps(nblocks) - blkweights_ps(pstart), minp1load)
      else
        ! hit
        mybottleneck = maxload
        myubload = maxload
      end if
      
      !write(*,'(I6,I6,E32.24,F7.4,L4)') mpi%rank,step,targetload,avload/targetload,sumload>=totalweight
      
      ! exit the loop if specified accuracy has been reached
      if(myubload-mylbload<=eps) then
        reqsteps = step + 1
        exit
      end if
      
    end do ! binary search steps
    
    !! every process has now its best found bottleneck value in variable mybottleneck
    !! if a process found no working bottleneck, this variable is equal ubload
    !! so the min value of all mybottleneck variables is the best found value
    
    call gettime_ofday(timing_us0)
    call MPI_Allreduce(mybottleneck, bottleneck, 1, MPI_R8K, MPI_MIN, mpi%comm, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allreduce failed')
    call gettime_ofday(mpi_time)
    mpi_time = mpi_time - timing_us0


    !if(mpi%rank==0) write(*,'(A,F7.4)') 'and the winner is:',avload/bottleneck

    !! finally create partition with given maximum load ("bottleneck value")

    guess = avgblk
    pstart = 0
    sumload = bottleneck
    do crank=1,nprocs-1
      ! find partition for rank crank with load <= bottleneck
#ifdef FD4_PART_1D_SCAN_HAN_BINSEARCH
      pend = fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, guess)
#else
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess)
#endif
      pstart = pend
      ! set partition vector
      partition(crank) = pstart+1
      sumload = bottleneck + blkweights_ps(pstart)
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck

  end subroutine fd4_part_1d_parallel


  !! Calculate a partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - very simple heuristic
  !!  - fast
  !!
  !! This is simply the method described as Heuristic 1 in:
  !!
  !! Miguet, S. und J.-M. Pierson: Heuristics for 1D rectilinear partitioning as a low cost
  !! and high quality answer to dynamic load balancing. in: High-Performance Computing and
  !! Networking, LNCS, vol. 1225, pages 550-564. Springer, 1997.
  !!
  subroutine fd4_part_1d_simple_h1(nblocks, nprocs, blkweights_ps, partition, maxload)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller

    integer(i_k) :: crank, pstart, pend, guess, avgblk
    real(r8k) :: sumload, bottleneck

    maxload = 0.0_r8k
    avgblk = nblocks / nprocs
    guess = avgblk
    pstart = 0  ! last block of prev partition
    bottleneck = blkweights_ps(nblocks) / nprocs
    sumload = bottleneck
    do crank=1,nprocs-1
      ! find partition for rank crank with load <= bottleneck
#ifdef FD4_PART_1D_SCAN_HAN
      pend = fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, guess)
#else
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess) ! last block of current partition
#endif
      !pend = fd4_part_1d_scan_bin(nblocks, blkweights_ps, sumload)  ! naively search whole weight array
      !maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
      pstart = pend
      ! set partition vector
      partition(crank) = pstart+1
      sumload = bottleneck + sumload
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck
    !maxload = MAX(maxload, blkweights_ps(partition(nprocs)-1) - blkweights_ps(partition(nprocs-1)-1))

  end subroutine fd4_part_1d_simple_h1


  !! Calculate a partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - very simple heuristic
  !!  - fast
  !!
  !! This is simply the method described as Heuristic 2 in:
  !!
  !! Miguet, S. und J.-M. Pierson: Heuristics for 1D rectilinear partitioning as a low cost
  !! and high quality answer to dynamic load balancing. in: High-Performance Computing and
  !! Networking, LNCS, vol. 1225, pages 550-564. Springer, 1997.
  !!
  subroutine fd4_part_1d_simple_h2(nblocks, nprocs, blkweights_ps, partition, maxload)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller

    integer(i_k) :: crank, pstart, pend, guess, avgblk
    real(r8k) :: sumload, bottleneck

    maxload = 0.0_r8k
    avgblk = nblocks / nprocs
    guess = avgblk
    pstart = 0  ! last block of prev partition
    bottleneck = blkweights_ps(nblocks) / nprocs
    sumload = bottleneck
    do crank=1,nprocs-1
      ! find partition for rank crank with load <= bottleneck
#ifdef FD4_PART_1D_SCAN_HAN
      pend = fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, guess)
#else
      pend = fd4_part_1d_scan(nblocks, blkweights_ps, pstart, sumload, guess) ! last block of current partition
#endif
      !pend = fd4_part_1d_scan_bin(nblocks, blkweights_ps, pstart, nblocks, sumload)  ! naively search whole weight array
      if(pend < nblocks) then
        if(blkweights_ps(pend+1) - sumload < sumload - blkweights_ps(pend)) pend = pend + 1
      end if
      !maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
      pstart = pend
      ! set partition vector
      partition(crank) = pstart+1
      sumload = bottleneck + sumload
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck
    !maxload = MAX(maxload, blkweights_ps(partition(nprocs)-1) - blkweights_ps(partition(nprocs-1)-1))

  end subroutine fd4_part_1d_simple_h2


  !! Calculate a partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - very simple heuristic
  !!  - fast
  !!
  !! Similar to $fd4_part_1d_simple_h1, but starts at the end of the partition vector (right-to-left).
  !!
  subroutine fd4_part_1d_simple_h1_reverse(nblocks, nprocs, blkweights_ps, partition, maxload)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller

    integer(i_k) :: crank, pstart, pend, guess, avgblk
    real(r8k) :: sumload, bottleneck

    maxload = 0.0_r8k
    avgblk = nblocks / nprocs
    guess = nblocks - avgblk
    pend = nblocks  ! last block of current partition
    bottleneck = blkweights_ps(nblocks) / nprocs
    sumload = blkweights_ps(nblocks) - bottleneck
    do crank=nprocs-1,1,-1
      ! find partition for rank crank with load <= bottleneck
      pstart = fd4_part_1d_scan_han_reverse(nblocks, blkweights_ps, sumload, avgblk, guess)
      !! adding the H2 shift leads to exactly the same results as "forward" H2:
      !if(pstart > 0) then
      !  if(blkweights_ps(pstart) - sumload > sumload - blkweights_ps(pstart-1)) pstart = pstart - 1
      !end if
      !maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
      pend = pstart
      ! set partition vector
      partition(crank) = pend+1
      sumload = sumload - bottleneck
    end do ! crank

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck
    !maxload = MAX(maxload, blkweights_ps(partition(nprocs)-1) - blkweights_ps(partition(nprocs-1)-1))

  end subroutine fd4_part_1d_simple_h1_reverse
  

  !! Calculate a partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! Algorithm:
  !!  - completely serial
  !!  - simple recursive bisection heuristic
  !!  - works for arbitrary proc numbers, not only powers of two
  !!  - fast
  !!
  subroutine fd4_part_1d_simple_rb(nblocks, nprocs, blkweights_ps, partition, maxload)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out), optional :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller

    integer(i_k) :: i
    
    call fd4_part_1d_simple_rb_rec(0, nblocks, 0, nprocs-1)
    
    partition(nprocs) = nblocks + 1
    
    if(present(maxload)) then
      maxload = 0.0_r8k
      do i=0,nprocs-1
        maxload = MAX(blkweights_ps(partition(i+1)-1) - blkweights_ps(partition(i)-1), maxload)
      end do
    end if
    
    contains
    
    !! weight of first block is: blkweights_ps(pi0+1) - blkweights_ps(pi0)
    !!
    recursive subroutine fd4_part_1d_simple_rb_rec(pi0, pi1, ri0, ri1)
      integer(i_k), intent(in)  :: pi0                   !! start of search interval (exclusive)
      integer(i_k), intent(in)  :: pi1                   !! end of search interval   (inclusive)
      integer(i_k), intent(in)  :: ri0                   !! start of rank interval   (inclusive)
      integer(i_k), intent(in)  :: ri1                   !! end of rank interval     (inclusive)
      
      integer(i_k) :: rim, pim
      real(r8k) :: sumload
      
      ! first rank in 2nd part
      rim = ISHFT(ri0 + ri1 + 1, -1)
      
      ! load of first part
      sumload =          REAL(rim-ri0,r8k)/REAL(ri1-ri0+1,r8k)  * blkweights_ps(pi1) + &
                (1.0_r8k-REAL(rim-ri0,r8k)/REAL(ri1-ri0+1,r8k)) * blkweights_ps(pi0)

      ! search first block in 2nd part
      !   binary search over current part of weight array [pi0, pi1]
      pim = fd4_part_1d_scan_bin(nblocks, blkweights_ps, pi0, pi1, sumload)

      ! increase pim if difference of sumload and blkweight_ps at pim+1 is smaller (like in H2)
      if(pim<pi1) then
        if(blkweights_ps(pim+1)-sumload < sumload-blkweights_ps(pim)) pim = pim + 1      
      end if

#ifdef PART1DEBUG
      write (*,'(I6,I6,I4,I4,A,I6,I4)') pi0,pi1,ri0,ri1,' : ',pim,ISHFT(ri0 + ri1 + 1, -1)
#endif

      if(ri1-ri0>1) then
        !! go further in recursion

        ! left  part: blocks [pi0+1, pim]  ranks [ri0,  rim-1]
        if(ri0<rim-1) then
          call fd4_part_1d_simple_rb_rec(pi0, pim, ri0, rim-1)
        else
          partition(ri0) = pi0 + 1
        end if

        ! right part: block [pim+1, pi1]   ranks [rim,  ri1]
        call fd4_part_1d_simple_rb_rec(pim, pi1, rim, ri1)

      else
        !! only two ranks in current part: we can now fill the partition vector and exit recursion
        
        partition(ri0) = pi0 + 1
        partition(ri1) = pim + 1
      
      end if
      
    end subroutine fd4_part_1d_simple_rb_rec

  end subroutine fd4_part_1d_simple_rb


  !! Calculate a well-balanced partition for given block weights.
  !! Note: weights have to be given as prefix sums!
  !!
  !! This is the method EBS (exact bisection) from:
  !!
  !! Pinar, A. and C. Aykanat: Fast optimal load balancing algorithms for 1D partitioning.
  !! Journal of Parallel and Distributed Computing, 64(8):974-996, 2004.
  !!
  subroutine fd4_part_1d_binsrch_ebs(nblocks, nprocs, blkweights_ps, maxweight, eps, maxsteps, cwsb, partition, maxload, reqsteps)
    integer(i_k), intent(in)  :: nblocks                  !! number of blocks
    integer(i_k), intent(in)  :: nprocs                   !! number of processes
    real(r8k),    intent(in)  :: blkweights_ps(0:nblocks) !! prefix sum of ordered block weights, weights(0) must be 0
    real(r8k),    intent(in)  :: maxweight                !! max weight in weight array
    real(r8k),    intent(in)  :: eps                      !! max. difference of the resulting bottleneck value from the minimal
    integer(i_k), intent(in)  :: maxsteps                 !! max. number of binary search steps, may stop earlier depending on eps
    logical,      intent(in)  :: cwsb                     !! construct final part. with separator bounds (T) or Han's method (F)?
    integer(i_k), intent(out) :: partition(0:nprocs)      !! output partition vector, contains start indices of partitions
    real(r8k),    intent(out) :: maxload                  !! estimation of load of max loaded process, real value is a bit smaller
    integer(i_k), intent(out) :: reqsteps                 !! required number of search steps to reach the accuracy specified by eps
    target :: partition

    integer(i_k) :: crank, pstart, pend, step
    real(r8k) :: totalweight, targetload, lbload, ubload, avload, sumload, bottleneck
    integer(i_k) :: guess, avgblk
    real(r8k) :: minp1load ! smallest realizable bottleneck value greater then current targetload
    
    real(r8k) :: bottleneck_rb  ! bottleneck from recursive bisection heuristic
    integer(i_k), target :: sp1(0:nprocs), sp2(0:nprocs) ! separator index bounds (high/low) for each partition
    integer(i_k), pointer :: shp(:), slp(:) ! separator index bounds (high/low) for each partition
    integer(i_k), pointer :: sp(:), sp_tmp(:)
    real(r8k) :: last_targetload ! targetload from last binary search step
    
    shp => sp1
    slp => sp2
    sp => partition  ! use partition array as one of the shp/slp/sp arrays
    
    reqsteps = maxsteps
    totalweight = blkweights_ps(nblocks)
    maxload = totalweight
    
    avgblk = nblocks / nprocs
    avload = totalweight / nprocs   ! average proc load = optimal balance
    !write(*,'(A,E12.4,E12.4,A,F7.4,F7.4)') 'lb/ub: ',lbload,ubload,' ~ ',avload/lbload,avload/ubload
    
    !!! Stage 1: find good approximation for separator index bounds (slp, shp)

    !! call recursive bisection heuristic to find a hopefully good bottleneck value (bottleneck_rb)
    call fd4_part_1d_simple_rb(nblocks, nprocs, blkweights_ps, partition, bottleneck_rb)

    !! run LR-PROBE with bottleneck_rb to find upper bounds for sep. idx
    guess = avgblk
    pstart = 0
    sumload = bottleneck_rb
    do crank=1,nprocs-1
      pend = fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, guess)
      pstart = pend
      shp(crank) = pstart+1
      sumload = blkweights_ps(pstart) + bottleneck_rb
    end do ! crank
    shp(0) = 1
    shp(nprocs) = nblocks + 1

    !! run RL-PROBE with bottleneck_rb to find lower bounds for sep. idx
    guess = nblocks - avgblk
    pend = nblocks  ! last block of current partition
    sumload = blkweights_ps(nblocks) - bottleneck_rb
    do crank=nprocs-1,1,-1
      pstart = fd4_part_1d_scan_han_reverse(nblocks, blkweights_ps, sumload, avgblk, guess)
      pend = pstart
      slp(crank) = pend+1
      sumload = blkweights_ps(pstart) - bottleneck_rb
    end do ! crank
    slp(0) = 1
    slp(nprocs) = nblocks + 1
    
    !print *,'slp',slp
    !print *,'shp',shp

    !! run LR-PROBE with avload as bottleneck to refine lower bounds for sep. idx
    sumload = avload
    do crank=1,nprocs-1
      pend = fd4_part_1d_scan_bin(nblocks, blkweights_ps, slp(crank)-1, shp(crank), sumload)
      slp(crank) = max(slp(crank), pend+1)
      sumload = blkweights_ps(pend) + avload
    end do ! crank

    !! run RL-PROBE with avload as bottleneck to refine upper bounds for sep. idx
    sumload = blkweights_ps(nblocks) - avload
    do crank=nprocs-1,1,-1
      pend = fd4_part_1d_scan_bin_reverse(nblocks, blkweights_ps, slp(crank)-1, shp(crank), sumload)
      shp(crank) = min(shp(crank), pend+2) !!!!!ML
      sumload = blkweights_ps(pend) - avload
    end do ! crank
    
    !print *,'slp',slp
    !print *,'shp',shp
    
   
    !!! Stage 2: run bisection algorithm, dynamically update the separator index bounds
    
    !! find good approximation of minimal max proc load ("bottleneck value")
    lbload = MAX(avload, maxweight) ! lower bound for max proc load, best value that can be reached
    ubload = bottleneck_rb          ! upper bound for max proc load, worst value
    targetload = -1.0_r8k
    do step=1,maxsteps

      last_targetload = targetload
      targetload = lbload * 0.5_r8k + ubload * 0.5_r8k

      if(last_targetload == targetload) then
        ! parachute in case floating point precision issues led us here
#ifdef PART1DEBUG
        print *, 'last_targetload == targetload'
#endif
        reqsteps = step - 1
        exit
      end if
      
#ifdef PART1DEBUG
      print *, 'step', step, ubload, lbload, targetload
#endif

      sumload = targetload
      pstart = 0
      maxload = 0
      minp1load = totalweight
      do crank=1,nprocs-1
        ! find partition for rank crank with load <= targetload
        pend = fd4_part_1d_scan_bin(nblocks, blkweights_ps, slp(crank)-1, shp(crank), sumload)
        !pend = fd4_part_1d_scan_bin(nblocks, blkweights_ps, 0, nblocks, sumload)
        maxload = MAX(blkweights_ps(pend) - blkweights_ps(pstart), maxload)
#ifdef PART1DEBUG
!        print *, crank, slp(crank)-1, shp(crank), pend+1
#endif
        if(pend==nblocks) exit
        minp1load = MIN(blkweights_ps(pend+1) - blkweights_ps(pstart), minp1load)
        pstart = pend
        sp(crank) = pstart+1 ! keep start index in sp array
        sumload = targetload + blkweights_ps(pstart)
      end do ! crank
      
      maxload = MAX(blkweights_ps(nblocks) - blkweights_ps(pstart), maxload)
#ifdef PART1DEBUG
      print *, 'maxload =',maxload, 'lastload =', blkweights_ps(nblocks) - blkweights_ps(pstart), maxload<=targetload
#endif
 
      if(maxload>targetload) then
#ifdef PART1EBSVERBOSE
        write(*,'(I5,F18.5,F18.5,F18.5,F12.8,F12.8,F12.8,I3,F12.8)') &
           step, targetload, lbload, ubload, avload/targetload, avload/lbload, avload/ubload, 0, avload/maxload
#endif        
        ! miss
        !lbload = MIN(blkweights_ps(nblocks) - blkweights_ps(pstart), minp1load)
        lbload = MIN(maxload, minp1load)
        !lbload = targetload
#ifdef PART1DEBUG
        print *, 'lbload =',lbload, 'minp1load =', minp1load, ubload>maxload
#endif
        ubload = min(ubload,maxload) ! modify ubload if maxload is smaller than current ubload
        ! swap slp and sp pointers
        sp_tmp => slp
        slp => sp
        sp => sp_tmp
      else
#ifdef PART1EBSVERBOSE
        write(*,'(I5,F18.5,F18.5,F18.5,F12.8,F12.8,F12.8,I3,F12.8)') &
           step, targetload, lbload, ubload, avload/targetload, avload/lbload, avload/ubload, 1, avload/maxload
#endif        
        ! hit
        ubload = maxload
        ! swap shp and sp pointers
        sp_tmp => shp
        shp => sp
        sp => sp_tmp
      end if
      
      !write(*,'(I6,E32.24,F7.4)') step,targetload,avload/targetload

      ! exit the loop if specified accuracy has been reached
      if(ubload-lbload<=eps) then
        reqsteps = step
        exit
      end if
      
    end do ! steps
    
    bottleneck = ubload
    
    !print *,ubload,lbload,ubload-lbload

    !!! Stage 3: finally create partition with given maximum load ("bottleneck value")

    pstart = 0
    sumload = bottleneck
    if(cwsb) then
      do crank=1,nprocs-1
        ! find partition for rank crank with load <= bottleneck
        pend = fd4_part_1d_scan_bin(nblocks, blkweights_ps, slp(crank)-1, shp(crank), sumload)
        pstart = pend
        ! set partition vector
        partition(crank) = pstart+1
        sumload = bottleneck + blkweights_ps(pstart)
      end do ! crank
    else
      guess = avgblk
      do crank=1,nprocs-1
        ! find partition for rank crank with load <= bottleneck
        pend = fd4_part_1d_scan_han(nblocks, blkweights_ps, sumload, avgblk, guess)
        pstart = pend
        ! set partition vector
        partition(crank) = pstart+1
        sumload = bottleneck + blkweights_ps(pstart)
      end do ! crank
    end if

    partition(0) = 1
    partition(nprocs) = nblocks + 1

    maxload = bottleneck

  end subroutine fd4_part_1d_binsrch_ebs


end module fd4_part_1d_mod
