! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2014 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                        Variable Table                        !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Variable table - a table for definition of data fields.
!!
!! *TODO*:
!!  - better documentation of vartab
!!  - different number of ghost cells for each variable
!!
!! @author Matthias Lieber
module fd4_vartab_mod

  use fd4_globaldef_mod
  
  implicit none
  private
   
  ! public types and methods
  public :: fd4_vartab, fd4_varinfo
  public :: fd4_vartab_check, fd4_vartab_create_varinfo

  !! length of variable name string
  integer(i_k), parameter, public ::  FD4_VARNAME_LEN = 64
  
  !! discretization types for variables
  integer(i_k), parameter, public ::  FD4_CELLC = 0  !! cell-centered
  integer(i_k), parameter, public ::  FD4_FACEX = 1  !! face-centered in x
  integer(i_k), parameter, public ::  FD4_FACEY = 2  !! face-centered in y
  integer(i_k), parameter, public ::  FD4_FACEZ = 3  !! face-centered in z

  !! value for fd4_vartab%vthres meaning no threshold
  !real(r_k), parameter, public  :: FD4_NOTHRES = huge(FD4_NOTHRES)
  ! PGI needs this workaround via additional real_variable:
  real(r_k), parameter, private :: real_variable = 0_r_k
  real(r_k), parameter, public  :: FD4_NOTHRES = huge(real_variable)
  
  
  !! This type represents an entry in the variable table.
  !! 
  !! Currently, cell-centered and face-centered variables are supported.
  !!  - Cell-centered variables: set $facevar to $FD4_CELLC.
  !!    *Only those variables have ghost cells*.
  !!  - Face-centered variables: set $facevar to the dimension ($FD4_FACEX, $FD4_FACEY, $FD4_FACEZ) of the face normal.
  !!    Face variables will have no ghost cells, but an additional cell plane in the the specified dimension
  !!    in *each* block.
  !!
  !! Hint: FD4 can do some optimizations when all variables have the same $vnull.
  !!
  type fd4_vartab
    character(len=FD4_VARNAME_LEN) :: name                   !! descriptive name of the variable
    integer(i_k)                   :: nbins = 1              !! number of bins of the variable
    integer(i_k)                   :: nsteps = 1             !! number of time steps for this variable
    logical                        :: dynamic = .false.      !! allocate data field dynamical?
    real(r_k)                      :: vnull = 0.0_r_k        !! initialization value of this variable
    real(r_k)                      :: vthres = FD4_NOTHRES   !! threshold to trigger block creation/removal
    integer(i_k)                   :: facevar = FD4_CELLC    !! indicator whether this variable is a face variable
  end type fd4_vartab

  
  !! This type contains some global information about the whole variable table.
  !!
  !! Use $fd4_vartab_create_varinfo to create this info from the vartab.
  type fd4_varinfo
    integer(i_k), pointer :: fields(:,:,:)            !! start and end indices of static fields for $fd4_block
    integer(i_k) :: maxnsteps                         !! maximal number of time steps in vartab
    integer(i_k) :: field_len(0:3)                    !! sum of all bins of all variables for cell-centered and facevars
    logical :: all_vars_same_vnull                    !! true if all variables have the same $vnull
    logical :: vthres_mode                            !! true if any variable has a $vthres /= $FD4_NOTHRES
  end type fd4_varinfo
    

contains


  !! Create fd4_varinfo from vartab.
  subroutine fd4_vartab_create_varinfo(vartab, varinfo, err)
    
    implicit none

    type(fd4_vartab) , intent(in)    :: vartab(:)    !! The variable table
    type(fd4_varinfo), intent(inout) :: varinfo      !! The variable table info
    integer(i_k)     , intent(out)   :: err          !! error status, 0 if ok (always 0 at error level < L3)

    integer(i_k) :: i, st, pos1(0:3), pos2(0:3), f
    
    err = -1
    
    ! check if all variables hava the same vnull
    varinfo%all_vars_same_vnull = .true.
    if(size(vartab)>1) then
      do i=2,size(vartab)
        if(vartab(1)%vnull /= vartab(i)%vnull) then
          varinfo%all_vars_same_vnull = .false.
          exit
        end if
      end do
    end if
    
    ! check if any variable has vthres
    varinfo%vthres_mode = .false.
    do i=1,size(vartab)
      if(vartab(i)%vthres < FD4_NOTHRES) then
        varinfo%vthres_mode = .true.
        exit
      end if
    end do    

    ! find maximum value of nsteps (number of time steps) in the variable table
    ! and count bins of static data fields
    varinfo%maxnsteps = 0
    varinfo%field_len(:) = 0
    do i=1,size(vartab)
      varinfo%maxnsteps = max( vartab(i)%nsteps, varinfo%maxnsteps )
      f = vartab(i)%facevar
      if( .not. vartab(i)%dynamic ) & 
        varinfo%field_len(f) = varinfo%field_len(f) + vartab(i)%nbins * vartab(i)%nsteps
    end do

    ! allocate data field index array
    allocate( varinfo%fields(size(vartab), varinfo%maxnsteps, 2) , stat=err )
    FD4_L3_ASSERT(err/=0,'allocate varinfo%fields')
    varinfo%fields(:,:,:) = 0

    ! set varinfo%fields
    ! change order of the loops to change the order of (variable, step) in the actual data array
    pos1(:) = 1
    pos2(:) = 0
    ! loop over maximum number of time steps
    do st=1,varinfo%maxnsteps
      ! loop over variables
      do i=1,size(vartab)
        if( .not. vartab(i)%dynamic .and. st <= vartab(i)%nsteps ) then
          ! vartab(i) is static data field with >= st time steps
          f = vartab(i)%facevar
          pos2(f) = pos1(f) + vartab(i)%nbins - 1
          FD4_L3_ASSERT(pos2(f)>varinfo%field_len(f),'pos2>len at '//vartab(i)%name)
          varinfo%fields(i,st,1) = pos1(f)
          varinfo%fields(i,st,2) = pos2(f)
          !print *,pos1(f),pos2(f),len(f),vartab(i)%name
          pos1(f) = pos2(f) + 1
        end if
      end do
    end do

#if ( FD4_VERBOSE_LEVEL >= 3 )
    do f=0,3
      FD4_L3_ASSERT(pos2(f)/=varinfo%field_len(f),'pos2(f)/=field_len(f), f='//i2s(f))
    end do
#endif
    
    err = 0

 
  end subroutine fd4_vartab_create_varinfo
  
  

  !! Check the variable table.
  !!
  !! Returns when the first error is found.
  !! Checks for:
  !!  - $nbins < 1
  !!  - $nsteps < 1
  !!  - duplicate variable names
  !!  - value of $facevar
  !!  - no dynamic face variables allowed
  !!  - face var cannot have threshold
  subroutine fd4_vartab_check(vartab, err)
    
    implicit none
    
    type(fd4_vartab), intent(in)  :: vartab(:)   !! The variable table
    integer(i_k)    , intent(out) :: err         !! error status, 0 if table is ok, else index of failed variable
    
    integer(i_k) :: i, j
    
    err = -1

    ! loop over variables
    do i=1, size(vartab)

      ! check number of bins
      if( vartab(i)%nbins < 1 ) then
         err = i
         FD4_L1_ERROR('number of bins (nbins) < 1: '//vartab(i)%name)
      end if
      
      ! check for nsteps < 1
      if( vartab(i)%nsteps < 1 ) then
         err = i
         FD4_L1_ERROR('number of time steps (nsteps) < 1: '//vartab(i)%name)
      end if

      ! check for variables with the same name
      do j=i+1, size(vartab)
        if( vartab(i)%name == vartab(j)%name ) then
          err = j
          FD4_L1_ERROR('double definition of variable name '//vartab(j)%name)
        end if
      end do
      
      ! check for face variables
      if( vartab(i)%facevar < 0 .or. vartab(i)%facevar > 3 ) then
        err = i
        FD4_L1_ERROR('facevar is invalid: '//vartab(i)%name)
      end if

      ! check if face variable is dynamic (that's not implemented)
      if( vartab(i)%facevar > 0 .and. vartab(i)%facevar < 4 .and. vartab(i)%dynamic ) then
        err = i
        FD4_L1_ERROR('sorry, facevar cannot be dynamic: '//vartab(i)%name)
      end if

      ! check if face variable has a threshold (that's not implemented)
      if( vartab(i)%facevar > 0 .and. vartab(i)%facevar < 4 .and. vartab(i)%vthres/=FD4_NOTHRES ) then
        err = i
        FD4_L1_ERROR('sorry, facevar cannot have a threshold: '//vartab(i)%name)
      end if
      
    end do
       
    err = 0
  
  end subroutine fd4_vartab_check


end module fd4_vartab_mod
