! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2014 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                       3D Morton Curve                        !
!                                                              !
!--------------------------------------------------------------!

#ifdef SFC_ERRCHECK
#define PURE
#endif

#ifndef SFC_MAX_LEV
#define SFC_MAX_LEV 12
#endif

!! Simple 3D Morton curve subroutines.
!!
!! Uses simple bit-interleaving algorithm as described in
!! M.J. Aftosmis, M.J. Berger and S.M. Murman, "Applications of Space-Filling Curves to Cartesian Methods for CFD", 2004
!! (except for the subroutines with limited query region size)
!!
!! Routines accept position and morton-index starting at 1!
!!
!! @author Matthias Lieber
module morton_mod

  use kinds_mod
  implicit none
  private  
  
  public :: morton_c2i
  public :: morton_c2i_qsz, morton_i2c_qsz
  public :: morton_iter, morton_iter_init, morton_iter_set, morton_iter_next

  integer(i_k), parameter :: SFC_DIMS = 3      !! spatial dimensions of SFC

  !! Iterator to traverse the 3D Morton curve
  !!
  type morton_iter
    integer(i_k) :: idx = - 1
    integer(i_k) :: p(SFC_DIMS) = - 1
    integer(i_k) :: lev = -1
    integer(i_k) :: qsz(SFC_DIMS) = -1
    integer(i_k) :: len = -1
    integer(i_k) :: c(SFC_MAX_LEV) = -1
    integer(i_k) :: p0(SFC_DIMS,SFC_MAX_LEV) = 1
    integer(i_K) :: jump_level = SFC_MAX_LEV + 1
  end type morton_iter
  
contains

include 'util/math.incf'

!! Get index on Morton curve at given 3D coordinate (coordinates-to-index).
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine morton_c2i(lev, p, idx)

  integer(i_k), intent(in)  :: lev    !! level of the morton curve (1=2x2x2, 2=4x4x4, 3=8x8x8, etc)
  integer(i_k), intent(in)  :: p(3)   !! 3D coordinates [1 ... 2^$lev]
  integer(i_k), intent(out) :: idx    !! index on the 3D morton curve [1 ... (2^$lev)^3]
  
  integer(i_k) :: i, mask, shft(3), p0(3)
  
  idx=0
  
  mask=1
  shft(1)=0
  shft(2)=1
  shft(3)=2
  p0=p-1
  
  do i=1,lev
    ! using of ior instead +
    idx  = ior( ior( idx, ishft(iand(p0(1),mask),shft(1)) )  , & 
           ior( ishft(iand(p0(2),mask),shft(2)) , ishft(iand(p0(3),mask),shft(3)) ) )
    mask = ishft(mask,1)  ! mask=mask*2
    shft = shft + 2
  end do

  idx = idx + 1
  
end subroutine morton_c2i


!! Get index on curve at given 3D coordinate (coordinates-to-index).
!!
!! The size of the query region is restricted by $qsz for each dimension.
!! $qsz must be <= $lev**2 and $pos must be <= $qsz.
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine morton_c2i_qsz(lev, qsz, p, idx)

  implicit none

  integer(i_k), intent(in)  :: lev    !! level of the morton curve (1=2x2x2, 2=4x4x4, 3=8x8x8, etc)
  integer(i_k), intent(in)  :: qsz(3) !! query region size
  integer(i_k), intent(in)  :: p(3)   !! 3D coordinates [1 ... 2^$lev]
  integer(i_k), intent(out) :: idx    !! index on the 3D morton curve [1 ... (2^$lev)^3]
  
  integer(i_k) :: l,dim,rq,cp(3)
  integer(i_k) :: csz(3), iq, rsz(3)
    
  cp(:)=p(:)-1
  idx=1
  dim=2**lev

  csz = qsz

#ifdef SFC_ERRCHECK  
  ! very expensive check
  if(cp(1)>=qsz(1) .or. cp(1)<0 .or. cp(2)>=qsz(2) .or. cp(2)<0 .or. cp(3)>=qsz(3) .or. cp(3)<0) then
    print *,'morton_c2i: position out of bounds', cp
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension and length of refined octants
    dim=dim/2
    
    ! determine spatial octant and reduce spatial coordinates
    ! so that they are relative to refined octant
    ! reference octant:
    ! z=0 | z=1  
    ! 2 3 | 6 7
    ! 0 1 | 4 5
    rq=0
    if(cp(1)>=dim) then
      rq=ior(1,rq)
      cp(1)=cp(1)-dim
    end if
    if(cp(2)>=dim) then
      rq=ior(2,rq)
      cp(2)=cp(2)-dim
    end if
    if(cp(3)>=dim) then
      rq=ior(4,rq)
      cp(3)=cp(3)-dim
    end if
    
    ! add offset to index
    do iq=0,rq-1
      ! count number of cells per quadrant that come before the current quadrant q
      if(iand(iq,1)==0) then
        rsz(1) = min(dim, csz(1))
      else
        rsz(1) = max(0  , csz(1)-dim)
      end if
      if(iand(iq,2)==0) then
        rsz(2) = min(dim, csz(2))
      else
        rsz(2) = max(0  , csz(2)-dim)
      end if
      if(iand(iq,4)==0) then
        rsz(3) = min(dim, csz(3))
      else
        rsz(3) = max(0  , csz(3)-dim)
      end if
      idx = idx + PRODUCT(rsz)
    end do
    
    ! shrink the current query size (csz) relative to current quadrant
    if(iand(rq,1)>0) then
      csz(1) = csz(1) - dim
    else
      csz(1) = min(csz(1), dim)
    end if
    if(iand(rq,2)>0) then
      csz(2) = csz(2) - dim
    else
      csz(2) = min(csz(2), dim)
    end if
    if(iand(rq,4)>0) then
      csz(3) = csz(3) - dim
    else
      csz(3) = min(csz(3), dim)
    end if
    
  end do
  
end subroutine morton_c2i_qsz


!! Get 3D coordinate at given index on curve (index-to-coordinates).
!!
!! The size of the query region is restricted by $qsz for each dimension.
!! $qsz must be <= $lev**2 and $pos must be <= $qsz.
!!
!! Has logarithmic complexity: number of loop iterations = $lev.
!! Due to performance reasons, no checking of arguments is performend!
PURE subroutine morton_i2c_qsz(lev, qsz, idx, p, iter)

  implicit none

  integer(i_k), intent(in)  :: lev    !! level of the morton curve (1=2x2x2, 2=4x4x4, 3=8x8x8, etc)
  integer(i_k), intent(in)  :: qsz(3) !! query region size
  integer(i_k), intent(in)  :: idx    !! index on the 3D morton curve [1 ... (2^$lev)^3 ]
  integer(i_k), intent(out) :: p(3)   !! 3D coordinates [1 ... 2^$lev]
  type(morton_iter), optional, intent(inout) :: iter !! for internal use only

  integer(i_k) :: l,dim,q,cidx
  integer(i_k) :: csz(3), iq, rsz(3), newcidx
  
  p(:)=1
  cidx=idx-1
  dim=2**lev
  
  csz = qsz

#ifdef SFC_ERRCHECK  
  if(cidx>=PRODUCT(qsz) .or. cidx<0) then
    print *,'morton_i2c: cidx >= PRODUCT(qsz) or cidx < 0: ', cidx
    stop
  end if
#endif

  do l=lev,1,-1

    ! dimension of refined octants
    dim=dim/2
    
    ! decrease cidx by offset in current quadrant
    ! and determine sequence number q of octant in the curve primitive.
    q=7
    do iq=0,6
      ! count number of cells per quadrant that come before the current quadrant q
      if(iand(iq,1)==0) then
        rsz(1) = min(dim, csz(1))
      else
        rsz(1) = max(0  , csz(1)-dim)
      end if
      if(iand(iq,2)==0) then
        rsz(2) = min(dim, csz(2))
      else
        rsz(2) = max(0  , csz(2)-dim)
      end if
      if(iand(iq,4)==0) then
        rsz(3) = min(dim, csz(3))
      else
        rsz(3) = max(0  , csz(3)-dim)
      end if
      newcidx = cidx - PRODUCT(rsz)
      if(newcidx>=0) then
        cidx = newcidx
      else
        q=iq
        exit
      end if
    end do

    ! shrink the current query size (csz) relative to current quadrant
    ! and update coordinates
    if(iand(q,1)>0) then
      p(1)=p(1)+dim
      csz(1) = csz(1) - dim
    else
      csz(1) = min(csz(1), dim)
    end if
    if(iand(q,2)>0) then
      p(2)=p(2)+dim
      csz(2) = csz(2) - dim
    else
      csz(2) = min(csz(2), dim)
    end if
    if(iand(q,4)>0) then
      csz(3) = csz(3) - dim
      p(3)=p(3)+dim
    else
      csz(3) = min(csz(3), dim)
    end if
    
    if(present(iter)) then
      if(l>1) iter%p0(:,l-1)=p
      iter%c(l)=q
    end if

  end do
 
end subroutine morton_i2c_qsz


!! Initialize the iterator.
!!
!! The current index is $iter%$idx and the current 3D position $iter%$p.
!!
!!> call morton_iter_init(iter, lev, qsz)
!!  do while(iter%idx > 0)
!!    array(iter%p(1),iter%p(2)) = iter%idx
!!    call morton_iter_next(iter)
!!  end do
!!< 
!!
PURE subroutine morton_iter_init(iter, lev, opt_qsz)

  type(morton_iter), intent(inout) :: iter
  integer(i_k), intent(in) :: lev
  integer(i_k), intent(in), optional :: opt_qsz(SFC_DIMS)
 
  if(lev>SFC_MAX_LEV) then
    iter%idx = -1
#ifdef SFC_ERRCHECK  
    print *,'morton_iter_init: lev > SFC_MAX_LEV = ',SFC_MAX_LEV
#endif
    return
  end if
  
  iter%idx = 1
  iter%p = 1
  iter%c = 0
  iter%lev = lev
  if(present(opt_qsz)) then
    iter%qsz = opt_qsz
    ! define the level at which we start jumping depending on the smallest dimension of the query size
    iter%jump_level = MAX( log2i(MINVAL(iter%qsz)) - 2, 1)
  else
    iter%qsz = 2**lev
    iter%jump_level = SFC_MAX_LEV + 1
  end if
  iter%len = PRODUCT(iter%qsz)
  iter%p0 = 1
    
end subroutine morton_iter_init


!! Set the iterator to given index
!!
PURE subroutine morton_iter_set(iter, idx)
  type(morton_iter), intent(inout) :: iter
  integer(i_k), intent(in) :: idx
  
  iter%idx = idx
  call morton_i2c_qsz(iter%lev, iter%qsz, iter%idx, iter%p, iter)
  
end subroutine morton_iter_set


!! Iterate to next point on the Morton curve
!!
!! After reaching the end of the curve, $iter%$idx is set to 0.
!!
PURE subroutine morton_iter_next(iter)

  type(morton_iter), intent(inout) :: iter
  
  integer(i_k) :: lev, toplev
  logical :: outside_qsz
  
  iter%idx = iter%idx + 1

  if(iter%idx>iter%len) then
    ! end of iteration
    iter%idx = 0
    return
  end if

  do

    lev = 1
    iter%c(1) = iter%c(1) + 1

    do while(iter%c(lev)>7)
      ! go up one level
      iter%c(lev) = 0
      iter%c(lev+1) = iter%c(lev+1) + 1
      lev = lev + 1
    end do
    toplev = lev - 1
    do lev=toplev,1,-1
      ! go down till bottom level
      iter%p0(1,lev) = iter%p0(1,lev+1) + ISHFT(IAND(iter%c(lev+1),1), lev)
      iter%p0(2,lev) = iter%p0(2,lev+1) + ISHFT(IAND(iter%c(lev+1),2), lev-1)
      iter%p0(3,lev) = iter%p0(3,lev+1) + ISHFT(IAND(iter%c(lev+1),4), lev-2)
      !print *,' next level+',lev, iter%c(lev), '  p:', iter%p0(:,lev)
    end do
    
    iter%p(1) = iter%p0(1,1) +       IAND(iter%c(1),1)
    iter%p(2) = iter%p0(2,1) + ISHFT(IAND(iter%c(1),2), -1)
    iter%p(3) = iter%p0(3,1) + ISHFT(IAND(iter%c(1),4), -2)
    
    !print *,iter%p,' |',iter%idx,iter%c(1:iter%lev)
    
    outside_qsz = ANY(iter%p>iter%qsz)

    if(outside_qsz .and. toplev>iter%jump_level) then
    !if(.false.) then
      ! jump if new point out of query region
      call morton_i2c_qsz(iter%lev, iter%qsz, iter%idx, iter%p, iter)
      !print *,' jumping to ', iter%idx,' at',iter%p
      outside_qsz = .false.
    end if
    
    if(.not.outside_qsz) exit
    !iter%idx = iter%idx - 1 ; exit
    
  end do
  

end subroutine morton_iter_next


end module morton_mod
