! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2014 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                      Timing Utilities                        !
!                                                              !
!--------------------------------------------------------------!

!! Timing Utilities.
!!
!! The subroutines return the time in microseconds as 64 bit integer.
!! The following timers are used:
!!  - $gettimeofday
!!  - $getrusage ($ru_utime)
!!  - $PAPIF_get_virt_usec
!!
!! @author Matthias Lieber
module timing_mod

  use kinds_mod, only: i8k, i_k
  implicit none
  
  private

  integer(i8k), parameter, public :: TIMING_MOD_NO_PAPI = -400000
  logical, public :: TIMING_MOD_HAS_PAPI
#ifdef PAPI_TIMING
  include 'f90papi.h'
  parameter ( TIMING_MOD_HAS_PAPI = .true. )
#else
  parameter ( TIMING_MOD_HAS_PAPI = .false. )
#endif

  !! the timers supported by this module
  integer(i_k), parameter, public :: TIMING_MOD_GETTIMEOFDAY = 1
  integer(i_k), parameter, public :: TIMING_MOD_RUSAGE_UTIME = 2
  integer(i_k), parameter, public :: TIMING_MOD_PAPI_VIRT    = 3
  
  character(len=14), parameter, public :: TIMING_MOD_TIMERS(1:3) = &
                  (/ 'gettimeofday  ', 'rusage:utime  ', 'papi:virt_usec' /)

  logical :: timing_initialized = .false.

  public :: gettime_init, gettime, gettime_ofday, gettime_rusage, gettime_papi
  
  !! Get the number of microseconds since the Epoch as returned by $gettimeofday.
  !!
  !! This is actually a C routine located in $gettimec.$c.
  !! It calls $gettimeofday to retrieve the number of microseconds since the
  !! Epoch (00:00:00 UTC, January 1, 1970).
  !!
  interface gettime_ofday
    subroutine gettimec(t)
      use kinds_mod, only: i8k
      integer(i8k), intent(out) :: t
    end subroutine gettimec
  end interface gettime_ofday


  !! Get the user-time of the process as returned by $ru_utime of $getrusage.
  !!
  !! This is actually a C routine located in $gettimec.$c that calls $getrusage.
  !! Caution: the resolution of this timer may be very bad.
  !!
  interface gettime_rusage
    subroutine gettimec_rusage(t)
      use kinds_mod, only: i8k
      integer(i8k), intent(out) :: t
    end subroutine gettimec_rusage
  end interface gettime_rusage

contains

  !! Initialize the timing module.
  !!
  !! This subroutine calls $PAPIF_library_init if compiled with PAPI support and
  !! if it is necessary before calling $PAPIF_get_virt_usec.
  !!
  subroutine gettime_init()
#ifdef PAPI_TIMING
    integer(i8k) :: t
    integer :: c
    if(timing_initialized) return
    call PAPIF_get_virt_usec(t)
    if(t<0) then
      ! we must initialize PAPI
      c=PAPI_VER_CURRENT 
      call PAPIF_library_init(c)
    end if
    timing_initialized = .true.
#else
    continue
#endif
  end subroutine gettime_init

  !! Get the virtual user time of the process as returned by PAPI.
  !!
  !! Uses the PAPI function $PAPIF_get_virt_usec to get the
  !! the time the process was running in user-mode
  !! since some arbitrary starting point.
  !!
  !! Returns $TIMING_MOD_NO_PAPI if FD4 is compiled without PAPI.
  !! Returns a negative value when PAPI failed, which may happen if PAPI is not initialized.
  !! (In this case, you need to call $gettime_init.)
  !!
  subroutine gettime_papi(t)
    integer(i8k), intent(out) :: t
#ifdef PAPI_TIMING
    call PAPIF_get_virt_usec(t)
#else
    t=TIMING_MOD_NO_PAPI
#endif
  end subroutine gettime_papi
  

  !! Get timing using a specified timer.
  !!
  !! Calls $gettime_ofday, $gettime_rusage, or $gettime_papi when $timer
  !! is $TIMING_MOD_GETTIMEOFDAY, $TIMING_MOD_RUSAGE_UTIME, or $TIMING_MOD_PAPI_VIRT, respectively.
  !! If no or no valid $timer is specified, $gettime_ofday is called.
  !!
  subroutine gettime(t, timer)
    integer(i8k), intent(out) :: t
    integer(i_k), intent(in), optional :: timer
    if(present(timer)) then
      select case(timer)
        case (TIMING_MOD_RUSAGE_UTIME)
          call gettime_rusage(t)
        case (TIMING_MOD_PAPI_VIRT)
          call gettime_papi(t)
        case default
          call gettime_ofday(t)
      end select
    else
      call gettime_ofday(t)      
    end if
  end subroutine gettime

end module timing_mod
